package Taketori;
use strict;
use utf8;
######################################################################
# HTML/プレーンTEXTを縦書き表示HTMLに変換するモジュール
# Taketori.pm 2.5.0
# Copyright (C) 1999-2013 CMONOS (ENDO chihiro)
#
# support@cmonos.jp
# http://cmonos.jp/
# 最終更新日：2013-07-26
######################################################################
# 
# このサブルーチン集の使い方
# ---------------------------------------------------------------------------------------
#	(0)-0 オブジェクトを作る。
#		require CMONOS::CGI_Lib::Taketori;
#		obj = Taketori->new({
#			'id'				=> 'tategaki-1',	# この縦書き区画(<div>)のid
#			'class'				=> 'tategaki',		# この縦書き区画(<div>)のclass
#			'style'				=> '',				# この縦書き区画(<div>)のstyle
#			'text'				=> 1,				# 1=改行コードでも改行
#			'margin'			=> 7,				# 行間(px)
#			'colmargin'			=> 7,				# 段間隔(px)
#			'fontsize'			=> 14,				# 基準フォントサイズ(px)
#			'rubysize'			=> 6,				# ルビフォントサイズ(px)
#			'height'			=> 480,				# 縦書き区画(<div>)の高さ(px)
#			'width'				=> 640,				# 縦書き区画(<div>)の幅(px)
#			'rotatedisabled'	=> 1,				# 長音符を横書き長音符の回転で表示せず「｜」で代用する
#			'blockenabled'		=> 1,				# すべてのブロック要素を許可する(ブロックごとに改段します)
#			'clear'				=> 1,				# 1=文末で回り込みを解除する
#			'ltrnum'			=> 1,				# 1=数字は横書きにする
#			'css'				=> 1,				# 1=CSSを分離
#		});
# 
#	(0)-1 初期設定を変更する。
#		obj->set(
#			...設定は(0)-0に準じます...
#		);
# 					
#	(0)-1 初期設定をリセットする。
#		obj->clear(
#			...設定は(0)-0に準じます...
#		);
# 					
# ---------------------------------------------------------------------------------------
#	(1) tategaki = obj->tategaki(text,option);
#		text を縦書き表示HTMLに変換します。
# 			text		変換するHTML <* class="ltr"> 部分は横書きのままにします。
# 			option		設定ハッシュのリファレンス または 設定テキスト
# 
# ---------------------------------------------------------------------------------------
#	(2)-1 obj->text_to_tategaki(text,tag);
#		テキスト中の<!--tate--><!--/tate-->で囲まれた部分を縦書き表示に変換します。
# 			text			テキスト
# 			tag				1=タグ（<!--_tate-->tate<!--/_tate-->）を付ける
# 
# 			※	タグに設定値を埋め込むことができます。
# 			例）	<!--tate width="300" height="150"-->text<!--/tate-->
# 
#	(2)-2 obj->tategaki_to_text(text);
#		テキスト中の<!--_tate--><!--/_tate-->で囲まれた部分を横書きに戻します。
# 			text			テキスト
# 
######################################################################
#	(0)-0 オブジェクトを作る。
use Jcode;
sub jcode_convert($$$$;$) {
my($this,$str,$ocode,$icode,$was_entity) = @_;
my $str_ref = (ref $str) ? $str : \$str;
$ocode = 'utf8' if !$ocode;
if (!$icode || $ocode ne $icode) {
	if (defined $Jcode::FALLBACK) {
		if ($ocode eq 'utf8') {
			$$str_ref = Jcode->new($str_ref,$icode)->fallback(&Jcode::FB_XMLCREF)->utf8;
		} elsif ($ocode eq 'euc') {
			$$str_ref = Jcode->new($str_ref,$icode)->fallback(&Jcode::FB_XMLCREF)->euc;
		} elsif ($ocode eq 'jis') {
			$$str_ref = Jcode->new($str_ref,$icode)->fallback(&Jcode::FB_XMLCREF)->jis;
		} elsif ($ocode eq 'sjis') {
			$$str_ref = Jcode->new($str_ref,$icode)->fallback(&Jcode::FB_XMLCREF)->sjis;
		} elsif ($ocode eq 'ucs2') {
			$$str_ref = Jcode->new($str_ref,$icode)->fallback(&Jcode::FB_XMLCREF)->ucs2;
		} else {
			Jcode::convert($str_ref,$ocode,$icode);
		}
	} else {
		Jcode::convert($str_ref,$ocode,$icode);
	}
}
if ($ocode eq 'utf8') {
	utf8::decode($$str_ref) if !utf8::is_utf8($$str_ref);
} elsif (utf8::is_utf8($$str_ref)) {
	utf8::encode($$str_ref);
}
$$str_ref =~ s/\&\#([0-9]{4,}|x[a-fA-F0-9]+);/$this->_entity_to_jcode($1,$ocode)/seg if !$was_entity && $ocode ne 'ucs2';
return $$str_ref;
}

sub _entity_to_jcode ($$$) {
my($this,$code,$ocode) = @_;
utf8::encode($code) if utf8::is_utf8($code);
$code = sprintf("%X",$code) unless $code =~ s/^x//;
return $this->jcode_convert(pack("U",hex($code)),$ocode,'utf8',1);
}

sub ename_to_jname ($$) {
my($this,$ename) = @_;
return 'utf8' if !$ename;
$ename = lc $ename;
my %ename2j = (
	'shiftjis'		=>	'sjis',
	'x-sjis'		=>	'sjis',
	'shift_jis'		=>	'sjis',
	'euc-jp'		=>	'euc',
	'x-euc-jp'		=>	'euc',
	'7bit-jis'		=>	'jis',
	'x-jis'			=>	'jis',
	'iso-2022-jp'	=>	'jis',
	'utf-16be'		=>	'ucs2',
	'utf-8'			=>	'utf8',
);
if ($ename2j{$ename}) {
	return $ename2j{$ename};
} else {
	$ename =~ tr/-/_/;
	return $ename;
}
}

sub h_to_z ($$;$) {
my($this,$text,$ascii) = @_;
return '' if !defined $text || $text eq '';
utf8::decode($text) if !utf8::is_utf8($text);
if ($ascii) {
	$text =~ s/\x20/\x{3000}/gs;
	$text =~ s/([\x{0020}-\x{007e}])/pack("U",unpack("U",$1) + 65248)/ges;
}
$text =~ s/･/・/gs;
$text =~ s/ｦ/ヲ/gs;
$text =~ s/ｧ/ァ/gs;
$text =~ s/ｨ/ィ/gs;
$text =~ s/ｩ/ゥ/gs;
$text =~ s/ｪ/ェ/gs;
$text =~ s/ｫ/ォ/gs;
$text =~ s/ｬ/ャ/gs;
$text =~ s/ｭ/ュ/gs;
$text =~ s/ｮ/ョ/gs;
$text =~ s/ｯ/ッ/gs;
$text =~ s/ｰ/ー/gs;
$text =~ s/ｱ/ア/gs;
$text =~ s/ｲ/イ/gs;
$text =~ s/ｳﾞ/ヴ/gs;
$text =~ s/ｳ/ウ/gs;
$text =~ s/ｴ/エ/gs;
$text =~ s/ｵ/オ/gs;
$text =~ s/ｶﾞ/ガ/gs;
$text =~ s/ｶ/カ/gs;
$text =~ s/ｷﾞ/ギ/gs;
$text =~ s/ｷ/キ/gs;
$text =~ s/ｸﾞ/グ/gs;
$text =~ s/ｸ/ク/gs;
$text =~ s/ｹﾞ/ゲ/gs;
$text =~ s/ｹ/ケ/gs;
$text =~ s/ｺﾞ/ゴ/gs;
$text =~ s/ｺ/コ/gs;
$text =~ s/ｻﾞ/ザ/gs;
$text =~ s/ｻ/サ/gs;
$text =~ s/ｼﾞ/ジ/gs;
$text =~ s/ｼ/シ/gs;
$text =~ s/ｽﾞ/ズ/gs;
$text =~ s/ｽ/ス/gs;
$text =~ s/ｾﾞ/ゼ/gs;
$text =~ s/ｾ/セ/gs;
$text =~ s/ｿﾞ/ゾ/gs;
$text =~ s/ｿ/ソ/gs;
$text =~ s/ﾀﾞ/ダ/gs;
$text =~ s/ﾀ/タ/gs;
$text =~ s/ﾁﾞ/ヂ/gs;
$text =~ s/ﾁ/チ/gs;
$text =~ s/ﾂﾞ/ヅ/gs;
$text =~ s/ﾂ/ツ/gs;
$text =~ s/ﾃﾞ/デ/gs;
$text =~ s/ﾃ/テ/gs;
$text =~ s/ﾄﾞ/ド/gs;
$text =~ s/ﾄ/ト/gs;
$text =~ s/ﾅ/ナ/gs;
$text =~ s/ﾆ/ニ/gs;
$text =~ s/ﾇ/ヌ/gs;
$text =~ s/ﾈ/ネ/gs;
$text =~ s/ﾉ/ノ/gs;
$text =~ s/ﾊﾞ/バ/gs;
$text =~ s/ﾊﾟ/パ/gs;
$text =~ s/ﾊ/ハ/gs;
$text =~ s/ﾋﾞ/ビ/gs;
$text =~ s/ﾋﾟ/ピ/gs;
$text =~ s/ﾋ/ヒ/gs;
$text =~ s/ﾌﾞ/ブ/gs;
$text =~ s/ﾌﾟ/プ/gs;
$text =~ s/ﾌ/フ/gs;
$text =~ s/ﾍﾞ/ベ/gs;
$text =~ s/ﾍﾟ/ペ/gs;
$text =~ s/ﾍ/ヘ/gs;
$text =~ s/ﾎﾞ/ボ/gs;
$text =~ s/ﾎﾟ/ポ/gs;
$text =~ s/ﾎ/ホ/gs;
$text =~ s/ﾏ/マ/gs;
$text =~ s/ﾐ/ミ/gs;
$text =~ s/ﾑ/ム/gs;
$text =~ s/ﾒ/メ/gs;
$text =~ s/ﾓ/モ/gs;
$text =~ s/ﾔ/ヤ/gs;
$text =~ s/ﾕ/ユ/gs;
$text =~ s/ﾖ/ヨ/gs;
$text =~ s/ﾗ/ラ/gs;
$text =~ s/ﾘ/リ/gs;
$text =~ s/ﾙ/ル/gs;
$text =~ s/ﾚ/レ/gs;
$text =~ s/ﾛ/ロ/gs;
$text =~ s/ﾜ/ワ/gs;
$text =~ s/ﾝ/ン/gs;
return $text;
}

sub z_to_h ($$) {
my($this,$text) = @_;
return '' if !defined $text || $text eq '';
utf8::decode($text) if !utf8::is_utf8($text);
$text =~ s/([\x{ff00}-\x{ff5e}])/pack("U",unpack("U",$1) - 65248)/ges;
$text =~ s/\x{3000}/\x20/gs;
return $text;
}

sub new ($;$$) {
my($pkg,$hash,$attributes) = @_;
my $this = {};
bless $this;
$this->set($hash);
$this->set($attributes);
$this;
}

#	(0)-1 初期設定
sub set ($$) {
my($this,$hash) = @_;
return $this if !defined $hash;
if (!ref $hash) {
	while ($hash =~ /[\x09\x0A\x0C\x0D\x20]*([0-9a-zA-Z_]+)[\x09\x0A\x0C\x0D\x20]*=[\x09\x0A\x0C\x0D\x20]*"([^"]*)"/gs) {
		$this->{lc($1)} = $this->unescape($2);
	}
} else {
	while (my($k,$v) = each %$hash) {
		next if !defined $k || $k eq '';
		$this->{lc($k)} = $v;
	}
}
$this->_setup;
return $this;
}

sub css ($) {
my $this = shift;
my $goma_margin = ($this->{'lang'} eq 'zh') ? '-1em -0.75em 0 0.75em' : '-1.3em -1em 0.3em 1em';
return <<CSS;
span.tktr-gyo {
  float: right !important;
  line-height: 1 !important;
  text-align: center !important;
  background: transparent !important;
  cursor: vertical-text;
}
span.tktr-gyo a {
  text-decoration: none !important;
  cursor: pointer !important;
}
span.tktr-gyo span.char {
  display: block !important;
}
span.tktr-gyo span {
  line-height: 1 !important;
}
span.tktr-gyo span.latin {
  display: block;
  margin: auto;
  -moz-transform: rotate(90deg);
  -webkit-transform: rotate(90deg);
  -o-transform: rotate(90deg);
  -ms-writing-mode: tb-rl;
  /writing-mode: tb-rl;
  _writing-mode: tb-rl;
}
span.tktr-gyo span.cho-on {
  display: block;
  margin: auto;
  -moz-transform: rotate(90deg) scale(1.0,-1.0);
  -webkit-transform: rotate(90deg) scale(1.0,-1.0);
  -o-transform: rotate(90deg) scale(1.0,-1.0);
  -ms-writing-mode: tb-rl;
  _writing-mode: tb-rl;
  /writing-mode: tb-rl;
}
span.tktr-gyo span.tou-ten {
  display: block;
  margin: -0.7em -0.64em 0.7em 0.64em;
  position:relative;
}
span.tktr-gyo span.ku-ten {
  display: block;
  margin: -0.7em -0.6em 0.7em 0.6em;
  position: relative;
}
span.tktr-gyo span.kogaki {
  display: block;
  margin: -0.2em -0.15em 0.2em 0.15em;
  position: relative
}
span.tktr-gyo strong {
  font-weight: normal;
  _border-right: 1px dotted;
  /border-right: 1px dotted;
}
span.tktr-gyo strong span:after {
  content: "\\3001";
  display: block;
  position: relative;
  width: 1px;
  height: 1em;
  margin: $goma_margin;
}
span.tktr-gyo strong span.latin,
span.tktr-gyo strong span.ltr {
  font-weight: bold;
  _border-right: none;
  /border-right: none;
}
span.tktr-gyo strong span.cho-on:after,
span.tktr-gyo strong span.latin:after,
span.tktr-gyo strong span.tou-ten:after,
span.tktr-gyo strong span.ku-ten:after,
span.tktr-gyo strong span.ltr:after {
  content: "";
  display: none;
}
span.tktr-gyo del {
  text-decoration: none;
  _text-decoration: line-through;
  /text-decoration: line-through;
}
span.tktr-gyo del span:after {
  content: "\\2503";
  display: block;
  position: relative;
  width: 1px;
  height: 1em;
  margin: -1em 0 0 0;
}
span.tktr-gyo del span.latin,
span.tktr-gyo del span.ltr {
  text-decoration: line-through;
}
span.tktr-gyo del span.cho-on:after,
span.tktr-gyo del span.latin:after,
span.tktr-gyo del span.tou-ten:after,
span.tktr-gyo del span.ku-ten:after,
span.tktr-gyo del span.ltr:after {
  content: "";
  display: none;
}
CSS
}

sub clear ($$) {
my($this,$hash) = @_;
undef %{$this};
return $this->set($hash);
}

sub _setup ($) {
my $this = shift;
$this->{'lang'} = 'ja' if !$this->{'lang'} || $this->{'lang'} =~ /[^a-zA-Z0-9\-_]/;
$this->{'fontsize'} = 14 if !$this->{'fontsize'} || $this->{'fontsize'} =~ /[^0-9]/;
if (!$this->{'height'}) {
	$this->{'height'} = $this->{'fontsize'} * 21;
} else {
	$this->{'height'} = int($this->{'fontsize'} * $1) if $this->{'height'} =~ /^([0-9]+\.?[0-9]*)em$/;
	$this->{'height'} = $this->{'fontsize'} * 21 if $this->{'height'} =~ /[^0-9]/;
	$this->{'height'} = $this->{'fontsize'} * 6 if $this->{'height'} < $this->{'fontsize'} * 6;
}
$this->{'rubysize'} = int($this->{'fontsize'}/2) if !$this->{'rubysize'} || $this->{'rubysize'} =~ /[^0-9]/;
$this->{'margin'} = $this->{'rubysize'} + 1 if !$this->{'margin'} || $this->{'margin'} =~ /[^0-9]/;
$this->{'colmargin'} = $this->{'fontsize'} if !defined $this->{'colmargin'} || $this->{'colmargin'} eq '' || $this->{'colmargin'} =~ /[^0-9]/;
$this->{'style'} .= ';' if $this->{'style'} && $this->{'style'} !~ /;$/;
}

#	(0)-3 設定の取得
sub attributes ($) {
my $this = shift;
return $this->hash_to_attributes($this);
}

sub hash_to_attributes ($) {
my($this,$hash) = @_;
my $attributes = '';
while (my($k,$v) = each %{$hash}) {
	next if !defined $k || !defined $v || ref $v || $k eq '' || $v eq '' || $k eq 'text';
	$attributes .= ' ' . $k . '="' . $this->escape($v) . '"';
}
return $attributes;
}

sub attributes_to_hash ($) {
my($this,$attributes) = @_;
return () if !$attributes;
my %hash = ();
while ($attributes =~ /[\x09\x0A\x0C\x0D\x20]*([0-9a-zA-Z_]+)[\x09\x0A\x0C\x0D\x20]*=[\x09\x0A\x0C\x0D\x20]*"([^"]*)"/gs) {
	$hash{$1} = $this->unescape($2);
}
return %hash;
}

sub escape ($$;$) {
my($this,$v,$escape_all) = @_;
return '' if !defined $v;
if ($escape_all) {
	$v =~ s/\&/&amp;/sg;
} else {
	$v =~ s/\&(?!(?:#[0-9]{2,6}|#x[a-fA-F0-9]+|[a-zA-Z0-9]{2,10});)/&amp;/sg;
}
$v =~ s/</&lt;/sg;
$v =~ s/>/&gt;/sg;
$v =~ s/"/&quot;/sg;
$v =~ s/\x0A/&#10;/g;
return $v;
}

sub unescape ($$) {
my($this,$v) = @_;
return '' if !defined $v;
$v =~ s/&lt;/</sg;
$v =~ s/&gt;/>/sg;
$v =~ s/&quot;/"/sg;
$v =~ s/&#10;/\x0A/sg;
$v =~ s/&amp;/\&/sg;
return $v;
}

######################################################################
#	(1) obj->tategaki(text[,option]);
#		textを縦書き表示HTMLに変換します。
sub tategaki ($$;$) {
my($this,$text,$option) = @_;
my $whole_html;
$this->set($option);
local $this->{'process'} = {
	'fontsize' => [$this->{'fontsize'}],
	'line_height' => [$this->{'height'}],
	'colmargin' => $this->{'colmargin'},
};

# utf8/LF形式に変換
if (utf8::is_utf8($text)) {
	$this->{'process'}->{'jcode'} = 'utf8';
} else {
	my $jcode = Jcode->new(\$text);
	$this->{'process'}->{'jcode'} = $this->ename_to_jname($jcode->icode);
	if (defined $Jcode::FALLBACK) {
		$text = $jcode->fallback(&Jcode::FB_XMLCREF)->utf8;
	} else {
		$text = $jcode->utf8;
	}
	utf8::decode($text);
}
$text = $this->h_to_z($text);
$text =~ s/\x0D\x0A/\x0A/gs;
$text =~ s/\x0D/\x0A/gs;
$text =~ s/\x09/\x20/gs;

#コメント削除
$text = $this->tategaki_to_text($text);
$text =~ s/<\!--tate(?![0-9a-zA-Z_])([^>]*)-->/<div class="ltr" style="width:$this->{'height'}px">/gs;
$text =~ s/<\!--\/tate-->/<\/div>/gs;
$text =~ s/<\!--(?=\/?[0-9a-zA-Z_]+).*?-->//gs;

# bodyあり
if ($text =~ /<body(?![0-9a-zA-Z_])[^>]*>/is) {
	$whole_html = $text;
	$text = $2 if $whole_html =~ s/(<body(?![0-9a-zA-Z_])[^>]*>)(.*)(<\/body>)/$1$3/is;
}

# ルビの調整
$text =~ s/<ruby(?![0-9a-zA-Z_])[^>]*>(.*?)<\/ruby>/$this->ruby_to_title($1)/egis;

# 縦中横
$text =~ s/(<[^>]+>|(\G|&\#[0-9]{2,6};|&\#x[a-fA-F0-9]+;|&[a-zA-Z]+;|[(\[\x{1100}-\x{11FF}\x{2030}-\x{217F}\x{2460}-\x{24FF}\x{2600}-\x{261B}\x{2620}-\x{277F}\x{2E80}-\x{2FDF}\x{2FF0}-\x{4DBF}\x{4E00}-\x{9FFF}\x{A960}-\x{A97F}\x{AC00}-\x{D7AF}\x{D7B0}-\x{D7FF}\x{F900}-\x{FAFF}\x{FE30}-\x{FE4F}\x{FF00}\x{FF01}\x{FF03}-\x{FF06}\x{FF08}-\x{FF0C}\x{FF0E}\x{FF0F}\x{FF1A}\x{FF1B}\x{FF1F}-\x{FF3D}\x{FF40}-\x{FF5B}\x{FF5D}-\x{FFEF}])[\x09\x0A\x0C\x0D\x20]*(\d{1,2}|[a-zA-Z]{1,2}|[\!\?]{1,2})(?=[\x09\x0A\x0C\x0D\x20]*(?:[\)\]<\x{1100}-\x{11FF}\x{2030}-\x{217F}\x{2460}-\x{24FF}\x{2600}-\x{261B}\x{2620}-\x{277F}\x{2E80}-\x{2FDF}\x{2FF0}-\x{4DBF}\x{4E00}-\x{9FFF}\x{A960}-\x{A97F}\x{AC00}-\x{D7AF}\x{D7B0}-\x{D7FF}\x{F900}-\x{FAFF}\x{FE30}-\x{FE4F}\x{FF00}\x{FF01}\x{FF03}-\x{FF06}\x{FF08}-\x{FF0C}\x{FF0E}\x{FF0F}\x{FF1A}\x{FF1B}\x{FF1F}-\x{FF3D}\x{FF40}-\x{FF5B}\x{FF5D}-\x{FFEF}]|$))[\x09\x0A\x0C\x0D\x20]*)/(($3) ? $2 . '<span class="ltr">' . $this->z_to_h($3) . '<\/span>' : $1)/ges;

# く の字点
$text =~ s/／＼/〳〵/gs;
$text =~ s/＼／/〳〵/gs;
$text =~ s/／["″]＼/〴〵/gs;
$text =~ s/＼["″]／/〴〵/gs;

# クォーテーション
$text =~ s/(<[^>]+>|\x20"([^"]*[^\x20-\x7F][^"]*)"\x20)/(($2) ? '『' . $2 . '』' : $1)/ges;

# 縦書き化（数字は横書き）
# タグ→エスケープ文字→エスケープ文字→2バイト文字（→桁数字）→1バイト文字→スペース→改行文字（3）
my($num,$tag,$entity,$abc,$crlf);
$num =  ($this->{'ltrnum'}) ? '|[\+\-]?[0-9][\,\.0-9]*' : '';
$tag = '<!--.*?-->|<[^>]+>';
$entity = '|&#[0-9]{2,6};|&#x[a-fA-F0-9]+;|&[a-zA-Z]+;';
$abc = '|[\?\!]{1,3}|[\x21-\x7f]|[\x20]+';
$crlf = '|\x0A';
$text =~ s/($tag$entity$num$abc$crlf|.)/$this->yoko_to_tate($1)/egs;
$text .= $this->line_close;

# 縦書きエリアを囲む
if ($this->{'id'} || $this->{'class'} || $this->{'style'} || $this->{'width'} || $this->{'wrapper'}) {
	$text = '<div'
		. (($this->{'id'}) ? ' id="' . $this->escape($this->{'id'}) . '"' : '')
		. (($this->{'class'}) ? ' class="' . $this->escape($this->{'class'}) . '"' : '')
		. (($this->{'style'} || $this->{'width'}) ? ' style="'
			. (($this->{'style'}) ? $this->escape($this->{'style'}) : '')
			. (($this->{'width'}) ? 'width:' . $this->{'width'} . 'px;' : '')
			. '"' : '')
		. '>'
		. $text
		. (($this->{'clear'}) ? '<span style="display:block;clear:right"></span>' : '')
		. '</div>';
} elsif ($this->{'clear'}) {
	$text .= '<span style="display:block;clear:right"></span>';
}

$text = $whole_html if $whole_html && $whole_html =~ s/(<body(?![0-9a-zA-Z_])[^>]*>)(<\/body>)/$1$text$2/is;
$this->jcode_convert(\$text,$this->{'process'}->{'jcode'},'utf8') if $this->{'process'}->{'jcode'} ne 'utf8';
return $text;
}

sub ruby_to_title ($$) {
my($this,$ruby) = @_;
my(@rb,@rt);
while ($ruby =~ /<rt>(.*?.)(?=(?:<\/?r[btp]>|$))/sgi) {
	push(@rt,$1);
}
if (!@rt) {
	$ruby =~ s/<\/?r[btp]>//sgi;
	return $ruby;
}
$ruby =~ s/<r[tp]>(.*?.)(?=(?:<\/?r[tp]>|$))/\0/sgi;
$ruby =~ s/<\/?r[btp]>/\0/sgi;
my @r = grep($_ ne '',split(/\0+/,$ruby));
push(@rb,@r) if @r;
if (@rb < @rt) {
	my $c = @rb;
	my @over = splice(@rt,$c,(@rt-$c));
	$rt[$#rt] .= join('',@over);
}
$ruby = '';
foreach my $i (0..$#rb) {
	if (defined $rt[$i] && $rt[$i] ne '') {
		my(@rbw,@rtw,$tag);
		$rt[$i] =~ s/<.*?>//sg;
		while ($rb[$i] =~ /(<.*?>|&\#[0-9]{2,6};|&\#x[a-fA-F0-9]+;|&[a-zA-Z]+;|[\?\!]{1,3}|.)/sg) {
			push(@rbw,$1);
		}
		while ($rt[$i] =~ /(&\#[0-9]{2,6};|&\#x[a-fA-F0-9]+;|&[a-zA-Z]+;|[\?\!]{1,3}|.)/sg) {
			push(@rtw,$1);
		}
		while (my $w = shift @rbw) {
			my $t = '';
			if (!@rtw || $w =~ /^<.*>$/s) {
				$ruby .= $w;
			} else {
				if (!@rbw) {
					$t = join('',@rtw);
				} elsif ((@rtw / (1 + @rbw)) > 1) {
					$t = join('',splice(@rtw,0,2));
				} else {
					$t = shift @rtw;
				}
				$ruby .= '<span title="' . $this->escape($t) . '" class="ruby-rb">' . $w . '</span>';
			}
		}
	} else {
		$ruby .= $rb[$i];
	}
}
return $ruby;
}

######################################################################
# 縦書き化
sub yoko_to_tate ($$) {
my($this,$w) = @_;
my $terminated = 0;
my $eject_out = 0;
my $floating_terminated = 0;
my $keep = 0;
my $text = '';

#テーブルタグ・横書き指定タグの中は処理しない
if (!$this->{'process'}{'waiting_for'} && !$this->{'process'}{'block_ltr'} && !$this->{'process'}{'inline_ltr'}) {

	#タグの場合の処理
	if ($w =~ /^<(\/?)([0-9a-zA-Z_]+)(?![0-9a-zA-Z_])([^>]*)(\/?)>$/s) {
		my($closed,$tag,$attributes,$is_empty) = ((($1) ? 1 : 0),lc($2),$3,(($4) ? 1 : 0));

		#改行
		if ($tag eq 'br') {
			$w = (($this->{'process'}{'text'}) ? '<br style="display:none" />' : '<span style="display:block;margin:0 auto;width:' . $this->{'process'}{'fontsize'}[0] . 'px;"><br /></span>') . "\x0A";
			$terminated = 1;

		#インライン要素
		} elsif (grep($_ eq $tag,'a','abbr','acronym','b','bdo','big','cite','code','dfn','em','font','i','kbd','label','q','s','samp','small','span','strike','strong','sub','sup','tt','u','var','del','ins')) {

			#閉じる
			if ($closed) {
				$w = $this->inline_tag_fix(\$tag);
				$this->{'process'}{'open_tags'} =~ s/<$tag(?![0-9a-zA-Z_])[^>]*>((?:(?!<$tag(?![0-9a-zA-Z_])[^>]*>).)*)$/$1/is if $this->{'process'}{'open_tags'};
				$this->{'process'}{'close_tags'} =~ s!</$tag>!!is if $this->{'process'}{'close_tags'};
				shift @{$this->{'process'}{'fontsize'}} if @{$this->{'process'}{'fontsize'}} > 1;
				shift @{$this->{'process'}{'inline_margin'}};
				shift @{$this->{'process'}{'inline_width'}};
				if ($this->{'process'}{'rt'}) {
					$w = $this->{'process'}{'rt'} . $w;
					delete $this->{'process'}{'rt'};
				}

			#開く
			} else {
				($w,$text) = $this->inline_style_fix(\$tag,$attributes);
				if (!$this->{'process'}{'inline_ltr'}) {
					$this->{'process'}{'close_tags'} = '</' . $tag . '>' . (($this->{'process'}{'close_tags'}) ? $this->{'process'}{'close_tags'} : '');
					$this->{'process'}{'open_tags'} .= $w;
				}
			}

		#オブジェクト
		} elsif (grep($_ eq $tag,'img','object','embed','applet','video','audio','canvas')) {
			my %attribute = $this->attributes_to_hash($attributes);
			my %style;
			if ($attribute{'style'}) {
				%style = $this->style_to_hash($attribute{'style'},'position|top|right|bottom|left|text-align|line-height|max|min|font|color');
				$attribute{'style'} = '';
				if ($style{'width'} && $style{'width'} =~ /([0-9]+)px/) {
					$attribute{'width'} = $1;
				} elsif (defined $style{'width'}) {
					delete $style{'width'};
				}
				if ($style{'height'} && $style{'height'} =~ /([0-9]+)px/) {
					$attribute{'height'} = $1;
				} elsif (defined $style{'height'}) {
					delete $style{'height'};
				}
			}
			if ($attribute{'align'}) {
				$style{'float'} = $attribute{'align'} if !$style{'float'};
				delete $attribute{'align'};
			}
			delete $style{'float'} if $style{'float'} && (!$attribute{'width'} || !$attribute{'height'});
			if ($tag eq 'img' && !$style{'float'} && (($attribute{'class'} && $attribute{'class'} eq 'gaiji') || ($attribute{'width'} && $attribute{'width'} < ($this->{'fontsize'} * 3) && $attribute{'height'} && $attribute{'height'} < ($this->{'fontsize'} * 3)))) {
				$attribute{'width'} = $attribute{'height'} = $this->{'process'}{'fontsize'}[0] if $attribute{'class'} && $attribute{'class'} eq 'gaiji';
				my($total_margin_width,$total_margin_height) = $this->calculate_margin(\%style);
				$style{'width'} = $attribute{'width'} . 'px' if $attribute{'width'};
				$style{'height'} = $attribute{'height'} . 'px' if $attribute{'height'};
				delete $attribute{'gaiji'};
				$attribute{'style'} = $this->hash_to_style(\%style);
				$w = '<img' . $this->hash_to_attributes(\%attribute) . ' />';
				my $total_indent = $this->total($this->{'process'}{'indent'}) + (($this->{'process'}{'prev_total_inline_margin'}) ? $this->{'process'}{'prev_total_inline_margin'} : 0);
				if ($this->{'process'}{'height'} && $this->{'process'}{'height'} > ($this->{'process'}{'line_height'}[0] - $total_indent - ($this->{'process'}{'fontsize'}[0] * 2))) {
					$text = $this->line_close;
				}
				$this->{'process'}{'height'} += $total_margin_height + $attribute{'height'};
				$this->set_object_width($total_margin_width + $attribute{'width'});
			} else {
				my $hack = '';
				my $floating_block = '';
				my $wrapper_height = $this->{'process'}{'line_height'}[0];
				my $link = '';
				my $indent = 0;
				$text = $this->line_close(1);
				my($total_margin_width,$total_margin_height) = $this->calculate_margin(\%style);
				if ($style{'float'}) {
					my $line_height = $this->{'process'}{'line_height'}[0] - $total_margin_height - $attribute{'height'};
					if (($style{'float'} eq 'left' || $style{'float'} eq 'right') && $line_height > $this->{'process'}{'fontsize'}[0] * 4) {
						$this->{'process'}{'floating_maxwidth'} = $attribute{'width'} + $total_margin_width;
						if ($style{'float'} eq 'right') {
							unshift(@{$this->{'process'}{'indent'}},$this->{'process'}{'indent'}[0]);
							$floating_block = '<span style="float:right;margin:' . $line_height . 'px 0 0 -' . $this->{'process'}{'floating_maxwidth'} . 'px;text-align:right">';
						} else {
							$indent = $this->{'process'}{'indent'}[0] if ref $this->{'process'}{'indent'} && defined $this->{'process'}{'indent'}[0];
							unshift(@{$this->{'process'}{'indent'}},0);
							$line_height -= $indent;
							$floating_block = '<span style="display:block;text-align:right">';
						}
						delete $style{'float'};
						$hack = $this->{'process'}{'floating_maxwidth'} + $this->{'process'}{'fontsize'}[0] + $this->{'margin'};
						$hack = '_width:' . $hack . 'px;/width:' . $hack . 'px;_margin-left:-' . $this->{'process'}{'fontsize'}[0] . 'px;/margin-left:-' . $this->{'process'}{'fontsize'}[0] . 'px;';
						unshift(@{$this->{'process'}{'line_height'}},$line_height);
					} elsif ($style{'float'} eq 'center') {
						$style{'margin'} = $this->add_margin($style{'margin'},int(($line_height - $this->{'process'}{'fontsize'}[0])/2),0,0,0);
					}
				} elsif (ref $this->{'process'}{'indent'} && @{$this->{'process'}{'indent'}}) {
					$indent = $this->{'process'}{'indent'}[0];
				}
				if (!$this->{'process'}{'floating_maxwidth'}) {
					delete $style{'float'};
					$style{'display'} = 'block';
				}
				if ($tag ne 'img' && !$is_empty) {
					$this->{'process'}{'waiting_for'} = $tag;
					$this->{'process'}{'waiting_tag'} = '</span>';
				}
				if (!$attribute{'height'} || !$attribute{'width'}) {
					$style{'max-height'} = $wrapper_height . 'px';
				} elsif ($attribute{'height'} > $wrapper_height) {
					$attribute{'width'} = int($attribute{'width'} * ($wrapper_height / $attribute{'height'}));
					$style{'width'} = $attribute{'width'} . 'px';
					$attribute{'height'} = $wrapper_height;
					$style{'height'} = $wrapper_height . 'px';
				}
				$attribute{'style'} = $this->hash_to_style(\%style);
				$link = $1 if $this->{'process'}{'prev_open_tags'} && $this->{'process'}{'prev_open_tags'} =~ /^.*(<a(?![0-9a-zA-Z_])[^>]*>)/is;
				$w = '<span style="'
					 . 'float:right;'
					 . 'height:' . $wrapper_height . 'px;'
					 . 'margin:0 ' . (($this->{'process'}{'margin_right'}) ? $this->{'process'}{'margin_right'} : '0') . ' ' . (($this->{'process'}{'colmargin'}) ? $this->{'process'}{'colmargin'} . 'px' : '0') . ' ' . ((!$this->{'process'}{'floating_maxwidth'} && $this->{'margin'}) ? $this->{'margin'} . 'px' : '0') . ';'
					 . (($this->{'process'}{'floating_maxwidth'}) ? 'min-width:' . $this->{'process'}{'floating_maxwidth'} . 'px;' : '')
					 . $hack . '">'
					 . (($floating_block) ? $floating_block : '')
					 . (($indent) ? '<span style="display:block;margin-top:' . $indent . 'px;">' : '')
					 . $link
					 . '<' . $tag . $this->hash_to_attributes(\%attribute) . ((!$this->{'process'}{'waiting_for'}) ? ' /' : '') . '>'
					 . (($link) ? '</a>' : '')
					 . (($indent) ? '</span>' : '')
					 . ((!$this->{'process'}{'waiting_for'}) ? '</span>' : '');
				$this->{'process'}{'colmargin'} = 0 if $this->{'process'}{'floating_maxwidth'};
				delete $this->{'process'}{'margin_right'};
				$eject_out = 1;
			}

		#そのまま書き出す
		} elsif (grep($_ eq $tag,'script','style','map','textarea','select','pre')) {
			$this->{'process'}{'waiting_for'} = $tag if !$is_empty;

		#区切り線
		} elsif ($tag eq 'hr') {
			$text = $this->line_close(1);
			my %attribute = $this->attributes_to_hash($attributes);
			my %style = $this->style_to_hash($attribute{'style'},'position|top|right|bottom|left|text-align|line-height|max|min|font|color');
			if (!$style{'border-top'} && !$style{'border'}) {
				$style{'border-top'} = 'solid ' . (($attribute{'size'}) ? $attribute{'size'} : '1') . 'px';
				$style{'border-right'} = $style{'border-bottom'} = $style{'border-left'} = 'none';
				$style{'background-color'} = 'transparent';
			}
			my $width = (($style{'height'}) ? $style{'height'} : '');
			if ($attribute{'width'} && !$style{'width'}) {
				$style{'width'} = $attribute{'width'} . (($attribute{'width'} !~ /\%$/) ? 'px' : '');
			}
			my $height;
			if ($style{'width'} && $style{'width'} =~ /([0-9]+\.?[0-9]*)\%/) {
				$height = int($1 * $this->{'process'}{'line_height'}[0] / 100);
			} elsif ($style{'width'} && $style{'width'} =~ /([0-9]+)px/) {
				$height = $1;
			} else {
				$height = $this->{'process'}{'line_height'}[0] - $this->{'process'}{'fontsize'}[0];
			}
			$style{'height'} = $height . 'px';
			my($total_margin_width,$total_margin_height) = $this->calculate_margin(\%style,$attribute{'align'});
			$style{'width'} = ($width) ? $width : ($total_margin_width) ? $total_margin_width . 'px' : '1px';
			if (($height + $total_margin_height) > $this->{'process'}{'line_height'}[0]) {
				$style{'height'} = ($this->{'process'}{'line_height'}[0] - $total_margin_height) . 'px';
			}
			$style{'float'} = 'right';
			$attribute{'style'} = 'display:none';
			$w = (($style{'clear'} && $style{'clear'} ne 'none') ? '<div style="clear:right"></div>' : '') . '<div style="float:right;width:' . $style{'width'} . ';height:' . $this->{'process'}{'line_height'}[0] . 'px;margin:0 0 ' . (($this->{'process'}{'colmargin'}) ? $this->{'process'}{'colmargin'} . 'px' : '0') . ' 0;"><div style="' . $this->hash_to_style(\%style) . '"><hr' . $this->hash_to_attributes(\%attribute) . ' /></div></div>' . "\x0A";
			$eject_out = 1;
			
		#その他
		} else {
			my($is_block,$is_header,$remain,$ltr);
			if (grep($_ eq $tag,'table','form','iframe','noframes')) {
				$is_block = $ltr = 1;
			} elsif (grep($_ eq $tag,'h1','h2','h3','h4','h5','h6')) {
				$is_block = $remain = $is_header = 1;
			} elsif (grep($_ eq $tag,'address','blockquote','center','div','dd','dl','dt','fieldset','li','ol','p','ul')) {
				$is_block = 1;
				$remain = 1 if $this->{'blockenabled'};
				if ($tag eq 'center') {
					$tag = 'div';
					$attributes = ' style="text-align:center"';
				}
			}
			if ($is_block) {
				$eject_out = 1;

				#閉じる
				if ($closed) {
					if ($remain) {
						$text = $this->line_close(1) . "\x0A";
					} else {
						$w = '';
						if ($this->{'process'}{'text'}) {
							$this->{'process'}{'text'} .= '<br style="display:none" />';
							$text .= $this->line_close . "\x0A";
						}
					}
					if ($tag eq 'ol' || $tag eq 'ul') {
						shift @{$this->{'process'}{'list_type'}};
						shift @{$this->{'process'}{'list_count'}};
					}
					$this->block_close($remain);

				#開く
				} else {
					my $had_text = ($this->{'process'}{'text'}) ? 1 : 0;
					$text = $this->line_close(1) . "\x0A";
					my %attribute = $this->attributes_to_hash($attributes);
					my %style = ($attribute{'style'}) ? $this->style_to_hash($attribute{'style'},'position|top|right|bottom|left') : ();
					if ($ltr || ($style{'display'} && $style{'display'} eq 'none') || ($attribute{'class'} && $attribute{'class'} =~ /(?:^|[\x09\x0A\x0C\x0D\x20])ltr(?:$|[\x09\x0A\x0C\x0D\x20])/) || ($attribute{'dir'} && $attribute{'dir'} eq 'ltr')) {
						$this->{'process'}{'block_ltr'}++;
						$w = (($style{'clear'} && $style{'clear'} ne 'none') ? '<div style="clear:right"></div>' : '') . '<div style="float:right;height:' . $this->{'process'}{'line_height'}[0] . 'px;margin:0 0 ' . (($this->{'process'}{'colmargin'}) ? $this->{'process'}{'colmargin'} . 'px' : '0') . ' 0;">' . $w;
					} else {
						my($line_height,$indent,$marigin_right,$is_top);
						$indent = 0;
						$is_top = (!ref $this->{'process'}{'indent'} || !@{$this->{'process'}{'indent'}}) ? 1 : 0;
						if ($tag eq 'ol' || $tag eq 'ul') {
							unshift(@{$this->{'process'}{'list_type'}},(($style{'list-style-type'}) ? $style{'list-style-type'} : ($attribute{'type'}) ? $attribute{'type'} : ($tag eq 'ol') ? '1' : 'disc'));
							unshift(@{$this->{'process'}{'list_count'}},0);
							$style{'list-style-type'} = 'none';
						} elsif ($tag eq 'li') {
							$this->{'process'}{'list_marker_required'} = 1;
							$this->{'process'}{'list_count'}[0]++;
							$indent = int($this->{'fontsize'} * 1.5);
						} elsif ($tag eq 'dd') {
							$indent = int($this->{'fontsize'} * 1.5);
						} elsif ($tag eq 'blockquote') {
							$indent = $this->{'fontsize'};
						}
						if ($is_top && $remain) {
							$style{'margin-top'} = $this->{'fontsize'} . 'px' if !$style{'margin'} && !$style{'margin-top'};
						}
						if (
								defined $style{'margin'} || defined $style{'margin-left'} || defined $style{'padding'} || defined $style{'padding-left'}
							||	($this->{'process'}{'line_height'}[0] - $this->total($this->{'process'}{'indent'}) - $indent) < $this->{'process'}{'fontsize'}[0] * 4
						) {
							$indent = 0;
						}
						my($total_margin_width,$total_margin_height) = $this->calculate_margin(\%style);
						$style{'margin'} = $this->add_margin($style{'margin'},0,0,$this->{'process'}{'colmargin'},0) if $remain && $this->{'process'}{'colmargin'};
						$this->block_open($this->{'process'}{'line_height'}[0],$total_margin_height,$indent,\$style{'font-size'},$style{'text-align'},$remain);
						if ($remain) {
							$style{'float'} = 'right' if !$style{'display'} || $style{'display'} ne 'none';
							if ($is_header && !$style{'width'}) {
								$style{'/width'} = $style{'_width'} = ($this->{'margin'} * 2 + $this->{'process'}{'fontsize'}[0]) . 'px';
							}
							$attribute{'style'} = $this->hash_to_style(\%style);
							$w = (($is_header && $style{'clear'} && $style{'clear'} ne 'none') ? '<div style="clear:right"></div>' : '') . '<' . $tag . $this->hash_to_attributes(\%attribute) . '>';
							$this->{'process'}{'margin_right'} = $this->{'margin'} . 'px' if $tag ne 'div' && $tag ne 'dd';
						} else {
							$w = ($is_top || $had_text) ? '<br style="display:none" />' : '';
							$this->{'process'}{'margin_right'} = $this->{'fontsize'} . 'px' if $tag ne 'div' && $tag ne 'dd';
							if (%style && ($style{'font-size'} || $style{'color'} || $style{'text-shadow'})) {
								$this->{'process'}{'close_tags'} = '</span>';
								$this->{'process'}{'open_tags'} = $this->{'process'}{'prev_open_tags'} = '<span style="' . $this->hash_to_style({'color' => $style{'color'},'font-size' => $style{'font-size'},'text-shadow' => $style{'text-shadow'}}) . '">';
							}
							if ($style{'margin'} && ($style{'margin'} =~ /[^\x09\x0A\x0C\x0D\x20]+[\x09\x0A\x0C\x0D\x20]+([^\x09\x0A\x0C\x0D\x20]+)[\x09\x0A\x0C\x0D\x20]+[^\x09\x0A\x0C\x0D\x20]+[\x09\x0A\x0C\x0D\x20]+[^\x09\x0A\x0C\x0D\x20]+/ || $style{'margin'} =~ /[^\x09\x0A\x0C\x0D\x20]+[\x09\x0A\x0C\x0D\x20]+([^\x09\x0A\x0C\x0D\x20]+)[\x09\x0A\x0C\x0D\x20]+[^\x09\x0A\x0C\x0D\x20]+/ || $style{'margin'} =~ /[^\x09\x0A\x0C\x0D\x20]+[\x09\x0A\x0C\x0D\x20]+([^\x09\x0A\x0C\x0D\x20]+)/ || $style{'margin'} =~ /([^\x09\x0A\x0C\x0D\x20]+)/) && $1) {
								$this->{'process'}{'margin_right'} = $1;
							}
						}
					}
				}
			}
		}

	#コメント
	} elsif ($w =~ /^<!--.*?-->$/s) {	

	#テキストファイルの場合改行コードがあれば改行する
	} elsif ($w =~ /^[\x0D\x0A]+$/s) {
		if ($this->{'text'}) {
			$w = '<br style="display:block;width:' . $this->{'process'}{'fontsize'}[0] . 'px;height:1px" />' . "\x0A";
			$terminated = 1;
		}
	
	#空白の連続は無視する
	} elsif ($this->{'process'}{'spacer'} && $w =~ /^\x20+$/s) {
		$w = '';

	#文字ほか
	} else {

		#文字の場合の処理
		if ($w !~ /^[\x0A\x0D]+$/s) {
			$this->{'process'}{'spacer'} = 0;

			# 禁則文字チェック
			if ($w =~ /[,﹐，\.﹒．、﹑､。｡〜ー～∼\)\}\]〉》」』】〕〗〙〛ぁぃぅぇぉっゃゅょゎゕゖゝゞァィゥェォッャュョヮヵヶヽヾ〵]/) {
				$keep = 1;
			}

			# 縦書き記号化+改行追加
			$w = $this->to_tategaki($w,$this->{'process'}{'fontsize'}[0]);
		}

		#禁則処理
		my $total_indent = $this->total($this->{'process'}{'indent'}) + (($this->{'process'}{'prev_total_inline_margin'}) ? $this->{'process'}{'prev_total_inline_margin'} : 0);
		if (		
				($this->{'process'}{'height'} && !$keep && $this->{'process'}{'height'} > ($this->{'process'}{'line_height'}[0] - $total_indent - ($this->{'process'}{'fontsize'}[0] * 2)))
			||	($this->{'process'}{'height'} && $this->{'process'}{'height'} > ($this->{'process'}{'line_height'}[0] - $total_indent - $this->{'process'}{'fontsize'}[0]))
		) {
			$text = $this->line_close;
		}
		$this->{'process'}{'height'} += $this->{'process'}{'fontsize'}[0];

	}

# 横書き区間 - タグ
} elsif ($w =~ /^<(\/?)([0-9a-zA-Z_]+)(?![0-9a-zA-Z_])([^>]*)(\/?)>$/s) {
	my($closed,$tag,$attributes,$is_empty) = ((($1) ? 1 : 0),lc($2),$3,(($4) ? 1 : 0));

	# 特定閉じタグ待ち
	if ($this->{'process'}{'waiting_for'}) {
		if ($closed && $tag eq $this->{'process'}{'waiting_for'}) {
			$w .= $this->{'process'}{'waiting_tag'} if $this->{'process'}{'waiting_tag'};
			delete $this->{'process'}{'waiting_for'};
			delete $this->{'process'}{'waiting_tag'};
		}
		$eject_out = 1 if !$this->{'process'}{'text'};

	# インライン要素
	} elsif ($this->{'process'}{'inline_ltr'} && grep($_ eq $tag,'a','abbr','acronym','b','bdo','big','cite','code','dfn','em','font','i','kbd','label','q','s','samp','small','span','strike','strong','sub','sup','tt','u','var','del','ins')) {
		if ($closed) {
			$this->{'process'}{'inline_ltr'}--;
		} else {
			$this->{'process'}{'inline_ltr'}++;
		}

	# ブロック要素
	} elsif ($this->{'process'}{'block_ltr'} && grep($_ eq $tag
		,'table','form','iframe','noframes'
		,'h1','h2','h3','h4','h5','h6'
		,'address','blockquote','center','div','dd','dl','dt','fieldset','li','ol','p','ul'
	)) {
		if ($closed) {
			$this->{'process'}{'block_ltr'}--;
			$w .= '</div>' if !$this->{'process'}{'inline_ltr'};
		} else {
			$this->{'process'}{'block_ltr'}++;
		}
		$w .= "\x0A";
		$eject_out = 1;
	} elsif ($this->{'process'}{'block_ltr'}) {
		$eject_out = 1;
	}

# 横書き区間 - 文字その他
} elsif ($this->{'process'}{'block_ltr'}) {
	$eject_out = 1;
}

# 行の高さが最大高さ - 2文字分を超えたら改行する
if ($eject_out) {
	$text .= $w;
} else {
	$this->{'process'}{'text'} .= $w;
	$text .= $this->line_close if $terminated;
}

return $text;
}

sub total ($;$) {
my($this,$array) = @_;
my $total = 0;
if (defined $array && ref $array) {
	foreach my $n (@$array) {
		$total += $n;
	}
}
return $total;
}

sub set_object_width ($$) {
my($this,$width) = @_;
$this->{'process'}{'inline_maxwidth'} = $width if !$this->{'process'}{'inline_maxwidth'} || $width > $this->{'process'}{'inline_maxwidth'};
}

sub line_close ($;$) {
my($this,$floating_terminated) = @_;
my $text = '';
my $floating_closed = 0;
my $line_width = ($this->{'process'}{'inline_maxwidth'}) ? $this->{'process'}{'inline_maxwidth'} : $this->{'process'}{'fontsize'}[0];
if ($this->{'process'}{'prev_line_width'} && $this->{'process'}{'prev_line_width'} > $line_width) {
	$line_width = $this->{'process'}{'prev_line_width'};
}
$this->{'process'}{'prev_line_width'} = (ref $this->{'process'}{'inline_width'}) ? $this->{'process'}{'inline_width'}[0] : $this->{'process'}{'fontsize'}[0];

#回り込みの解除
if ($this->{'process'}{'floating_maxwidth'}) {
	$this->{'process'}{'floating_width'} += $this->{'margin'} + $line_width + (($this->{'process'}{'margin_right'} && $this->{'process'}{'margin_right'} =~ /^([0-9]+)/) ? $1 : 0);
	if ($floating_terminated || $this->{'process'}{'floating_width'} > $this->{'process'}{'floating_maxwidth'}) {
		delete $this->{'process'}{'floating_width'};
		delete $this->{'process'}{'floating_maxwidth'};
		$floating_closed = 1;
	}
}

if ($this->{'process'}{'text'} && ($this->{'process'}{'text'} =~ /^[\x09\x0A\x0C\x0D\x20]+$/s || $this->{'process'}{'text'} =~ /^[\x09\x0A\x0C\x0D\x20]*(?:<br style="display:none" \/>[\x09\x0A\x0C\x0D\x20]*)+$/is)) {
	$text = $this->{'process'}{'text'};
	
} elsif ($this->{'process'}{'text'} && $this->{'process'}{'text'} =~ /^(?:[\x09\x0A\x0C\x0D\x20]*<span(?![0-9a-zA-Z_])[^>]*>(?:[\x09\x0A\x0C\x0D\x20]*|&nbsp;)<\/span>)+[\x09\x0A\x0C\x0D\x20]*$/is) {
	$text = ' ';
	
} elsif ($this->{'process'}{'text'} && $this->{'process'}{'text'} !~ /^[\x09\x0A\x0C\x0D\x20]*<span(?![0-9a-zA-Z_])[^>]*><br \/><\/span>[\x09\x0A\x0C\x0D\x20]*$/is && $this->{'process'}{'text'} =~ /^(?:[\x09\x0A\x0C\x0D\x20]*<[^>]+>)+[\x09\x0A\x0C\x0D\x20]*$/is) {
	$text = '';
	
} elsif ($this->{'process'}{'text'}) {

	#インラインタグの継承
	if ($this->{'process'}{'prev_open_tags'} || $this->{'process'}{'close_tags'}) {
		my $close_tags = ($this->{'process'}{'close_tags'}) ? $this->{'process'}{'close_tags'} : '';
		if ($close_tags) {
			while ($this->{'process'}{'close_tags'} =~ /^<\/([0-9a-zA-Z_]+)>/) {
				if ($this->{'process'}{'text'} =~ s/<($1)(?![0-9a-zA-Z_])[^>]*>[\x09\x0A\x0C\x0D\x20]*$//s) {
					$close_tags =~ s/^<\/([0-9a-zA-Z_]+)>//;
				} else {
					last;
				}
			}
		}
		$this->{'process'}{'text'} = (($this->{'process'}{'prev_open_tags'}) ? $this->{'process'}{'prev_open_tags'} : '') . $this->{'process'}{'text'} . $close_tags;
	}

	#リストマーカー
	if ($this->{'process'}{'list_marker_required'}) {
		my $list_type = $this->{'process'}{'list_type'}[0] || '';
		my $n = $this->{'process'}{'list_count'}[0];
		my $marker = '';
		if ($n > 0) {
			if ($list_type =~ /^none$/i) {
				$marker = '';
			} elsif ($list_type =~ /^disc$/i) {
				$marker = "●";
			} elsif ($list_type =~ /^circle$/i) {
				$marker = "○";
			} elsif ($list_type =~ /^square$/i) {
				$marker = "□";		
			} elsif (($this->{'lang'} eq 'ja' && ($list_type eq 'A' || $list_type eq 'upper-latin')) || $list_type eq 'ア' || $list_type eq 'katakana') {
				my @list = ('ア','ア','イ','ウ','エ','オ','カ','キ','ク','ケ','コ','サ','シ','ス','セ','ソ','タ','チ','ツ','テ','ト','ナ','ニ','ヌ','ネ','ノ','ハ','ヒ','フ','ヘ','ホ','マ','ミ','ム','メ','モ','ヤ','ユ','ヨ','ラ','リ','ル','レ','ロ','ワ','ヰ','ヱ','ヲ','ン');
				$marker = '(' . $list[$n] . ')';
			} elsif ($list_type eq 'A' || $list_type eq 'upper-latin') {
				my @list = ('A','A'..'Z');
				$marker = '(' . $list[$n] . ')';
			} elsif (($this->{'lang'} eq 'ja' && ($list_type eq 'a' || $list_type eq 'lower-latin')) || $list_type eq 'あ' || $list_type eq 'hiragana') {
				my @list = ('あ','あ','い','う','え','お','か','き','く','け','こ','さ','し','す','せ','そ','た','ち','つ','て','と','な','に','ぬ','ね','の','は','ひ','ふ','へ','ほ','ま','み','む','め','も','や','ゆ','よ','ら','り','る','れ','ろ','わ','ゐ','ゑ','を','ん');
				$marker = '(' . $list[$n] . ')';
			} elsif ($list_type eq 'a' || $list_type eq 'lower-latin') {
				my @list = ('a','a'..'z');
				$marker = '(' . $list[$n] . ')';
			} elsif ($list_type eq 'イ' || $list_type eq 'katakana-iroha') {
				my @list = ('イ','イ','ロ','ハ','ニ','ホ','ヘ','ト','チ','リ','ヌ','ル','ヲ','ワ','カ','ヨ','タ','レ','ソ','ツ','ネ','ナ','ラ','ム','ウ','ヰ','ノ','オ','ク','ヤ','マ','ケ','フ','コ','エ','テ','ア','サ','キ','ユ','メ','ミ','シ','ヱ','ヒ','モ','セ','ス');
				$marker = '(' . $list[$n] . ')';
			} elsif ($list_type eq 'い' || $list_type eq 'hiragana-iroha') {
				my @list = ('い','い','ろ','は','に','ほ','へ','と','ち','り','ぬ','る','を','わ','か','よ','た','れ','そ','つ','ね','な','ら','む','う','ゐ','の','お','く','や','ま','け','ふ','こ','え','て','あ','さ','き','ゆ','め','み','し','ゑ','ひ','も','せ','す');
				$marker = '(' . $list[$n] . ')';
			} elsif ($list_type eq 'I' || $list_type eq 'upper-roman') {
				my @list = ('I','I','II','III','IV','V','VI','VII','VIII','IX','X','XI','XII','XIII','XIV','XV','XVI','XVII','XVIII','XIX','XX'); 
				$marker = '(' . $list[$n] . ')';
			} elsif ($list_type eq 'i' || $list_type eq 'lower-roman') {
				my @list = ('i','i','ii','iii','iv','v','vi','vii','viii','ix','x','xi','xii','xiii','xiv','xv','xvi','xvii','xviii','xix','xx'); 
				$marker = '(' . $list[$n] . ')';
			} elsif ($list_type eq '1' || $list_type eq 'decimal' || $list_type eq '一' || $list_type eq 'cjk-ideographic' || $list_type eq '壱') {
				if ($n <= 10) {
					my @list = ($list_type eq '壱') ? ('壱','壱','弐','参','四','伍','六','七','八','九','拾') : ('一','一','二','三','四','五','六','七','八','九','十'); 
					$marker = '(' . $list[$n] . ')';
				} else {
					$marker = '(' . $n . ')';
				}
			} else {
				$marker = "・";
			}
			if ($marker) {
				$marker = '<span style="display:block;margin:-' . $this->{'fontsize'} . 'px auto 0 auto;font-size:' . $this->{'fontsize'} . 'px;white-space:nowrap;">' . $marker . '</span>';
				$this->{'process'}{'text'} = $marker . $this->{'process'}{'text'} if $this->{'process'}{'text'};
			}
		}
	}

	#垂直位置の調整
	my $margin_top = $this->total($this->{'process'}{'indent'});
	if ($this->{'process'}{'valign'}[0]) {
		$margin_top += $this->ratio_to_px('auto','',$this->{'process'}{'height'},(($this->{'process'}{'valign'}[0] && $this->{'process'}{'valign'}[0] eq 'center') ? 'auto' : 0),($this->{'process'}{'line_height'}[0] - $margin_top - $this->{'process'}{'fontsize'}[0]));
	}
	$this->{'process'}{'text'} = '<span style="display:block;margin-top:' . $margin_top . 'px;">' . $this->{'process'}{'text'} . '</span>' if $margin_top;

	#行を作る
	$text = '<span class="tktr-gyo" style="' . (($this->{'css'}) ? '' : 'float:right;line-height:1;text-align:center;background:transparent;') . 'height:' . $this->{'process'}{'line_height'}[0] . 'px;margin:0 ' . (($this->{'process'}{'margin_right'}) ? $this->{'process'}{'margin_right'} : '0') . ' ' . (($this->{'process'}{'colmargin'} && !$this->{'process'}{'floating_maxwidth'}) ? $this->{'process'}{'colmargin'} . 'px' : '0') . ' ' . (($this->{'margin'}) ? $this->{'margin'} . 'px' : '0') . ';font-size:' . $this->{'fontsize'} . 'px;_width:' . $line_width . 'px;/width:' . $line_width . 'px;">' . $this->{'process'}{'text'} . '</span>';
	delete $this->{'process'}{'margin_right'};
}

#回り込み終了
if ($floating_closed) {
	$text .= '</span>';
	shift @{$this->{'process'}{'line_height'}} if ref $this->{'process'}{'line_height'} && @{$this->{'process'}{'line_height'}};
	shift @{$this->{'process'}{'indent'}} if ref $this->{'process'}{'indent'} && @{$this->{'process'}{'indent'}};
	$this->{'process'}{'colmargin'} = $this->{'colmargin'} if !$this->{'process'}{'in_block'};
}
$this->{'process'}{'prev_open_tags'} = ($this->{'process'}{'open_tags'}) ? $this->{'process'}{'open_tags'} : '';
$this->{'process'}{'prev_total_inline_margin'} = $this->total($this->{'process'}{'inline_margin'});
delete $this->{'process'}{'text'};
delete $this->{'process'}{'height'};
delete $this->{'process'}{'inline_maxwidth'};
delete $this->{'process'}{'list_marker_required'};
return $text;
}

sub inline_tag_fix ($$) {
my($this,$tag) = @_;
$$tag =~ s/^(?:strike|s)$/del/ || $$tag =~ s/^b$/strong/ || $$tag =~ s/^i$/em/ || $$tag =~ s/^(?:font|big|small|tt|u)$/span/;
return '</' . $$tag . '>';
}

sub inline_style_fix ($$$) {
my($this,$tag,$attributes) = @_;
my %attribute = $this->attributes_to_hash($attributes);
my %style;
if ($attribute{'style'}) {
	%style = $this->style_to_hash($attribute{'style'},'position|top|right|bottom|left|float|text-align|line-height|width|height|max|min');
}
if (($attribute{'class'} && $attribute{'class'} =~ /(?:^|[\x09\x0A\x0C\x0D\x20])ltr(?:$|[\x09\x0A\x0C\x0D\x20])/) || ($style{'display'} && $style{'display'} eq 'none') || ($attribute{'dir'} && $attribute{'dir'} eq 'ltr')) {
	$this->{'process'}{'inline_ltr'}++;
	my $text = '';
	if (!$style{'display'} || $style{'display'} ne 'none') {
		my $total_indent = $this->total($this->{'process'}{'indent'}) + (($this->{'process'}{'prev_total_inline_margin'}) ? $this->{'process'}{'prev_total_inline_margin'} : 0);
		if ($this->{'process'}{'height'} && $this->{'process'}{'height'} > ($this->{'process'}{'line_height'}[0] - $total_indent - ($this->{'process'}{'fontsize'}[0] * 2))) {
			$text = $this->line_close;
		}
		$this->{'process'}{'height'} += $this->{'process'}{'fontsize'}[0];
	} elsif (!$attribute{'class'} || $attribute{'class'} !~ /(?:^|[\x09\x0A\x0C\x0D\x20])ltr(?:$|[\x09\x0A\x0C\x0D\x20])/) {
		$attribute{'class'} .= (($attribute{'class'}) ? ' ' : '') . 'ltr';
	}
	return('<' . $$tag . $attributes . '>',$text);
}
$$tag = lc $$tag;
$$tag =~ s/^(?:strike|s)$/del/ || $$tag =~ s/^b$/strong/ || $$tag =~ s/^i$/em/;
if ($$tag eq 'font') {
	$$tag = 'span';
	if ($attribute{'size'}) {
		if ($attribute{'size'} == 1) {
			$style{'font-size'} = '9px';
		} elsif ($attribute{'size'} == 2) {
			$style{'font-size'} = '10px';
		} elsif ($attribute{'size'} == 3) {
			$style{'font-size'} = '12px';
		} elsif ($attribute{'size'} == 4) {
			$style{'font-size'} = '14px';
		} elsif ($attribute{'size'} == 5) {
			$style{'font-size'} = '18px';
		} elsif ($attribute{'size'} == 6) {
			$style{'font-size'} = '24px';
		} elsif ($attribute{'size'} >= 7) {
			$style{'font-size'} = '36px';
		}
		delete $attribute{'size'};
	}
	if ($attribute{'face'}) {
		$style{'font-family'} = $attribute{'face'};
		delete $attribute{'face'};
	}
	if ($attribute{'color'}) {
		$style{'color'} = $attribute{'color'};
		delete $attribute{'color'};
	}
} elsif ($$tag eq 'big') {
	$$tag = 'span';
	$style{'font-size'} = '23px';
} elsif ($$tag eq 'small') {
	$$tag = 'span';
	$style{'font-size'} = '11px';
} elsif ($$tag eq 'tt') {
	$$tag = 'span';
	$style{'font-family'} = 'monospace';
} elsif ($$tag eq 'u') {
	$$tag = 'span';
	$style{'padding'} = '1px 0';
	$style{'border-top'} = '1px solid';
	$style{'text-decoration'} = 'none';
} elsif ($$tag eq 'ins') {
	$style{'padding'} = '1px 0';
	$style{'border-top'} = '1px dotted';
	$style{'text-decoration'} = 'none';
} elsif ($$tag eq 'em') {
	if (!$style{'border-top'}) {
		$style{'padding'} = '1px 0';
		$style{'border-top'} = '1px dotted';
	}
	$style{'font-style'} = 'normal' if !$style{'font-style'};
}
if ($$tag eq 'a' && !$style{'text-decoration'}) {
	$style{'text-decoration'} = 'none';
} elsif ($style{'text-decoration'}) {
	$style{'border-bottom'} = '1px solid' if $style{'text-decoration'} =~ s/[\x09\x0A\x0C\x0D\x20]*overline//ig;
	$style{'border-top'} = '1px solid' if $style{'text-decoration'} =~ s/[\x09\x0A\x0C\x0D\x20]*underline//ig;
	if ($style{'text-decoration'} !~ /[^\x09\x0A\x0C\x0D\x20]/) {
		$style{'padding'} = '1px 0';
		delete $style{'text-decoration'};
	}
}

$this->update_font_size(($style{'font-size'}) ? \$style{'font-size'} : '');
my($total_margin_width,$total_margin_height) = $this->calculate_margin(\%style);
$this->{'process'}{'height'} += $total_margin_height;
unshift(@{$this->{'process'}{'inline_margin'}},$total_margin_height);
unshift(@{$this->{'process'}{'inline_width'}},($total_margin_width + $this->{'process'}{'fontsize'}[0]));
$this->set_object_width($total_margin_width + $this->{'process'}{'fontsize'}[0]);

if ($attribute{'class'} && $attribute{'class'} =~ /(?:^|[\x09\x0A\x0C\x0D\x20])ruby-rb(?:$|[\x09\x0A\x0C\x0D\x20])/) {
	if ($attribute{'title'}) {
		my $count = ($attribute{'title'} =~ s/(&#x[a-fA-F0-9]+;|&#[0-9]{2,6};|&[a-zA-Z]+;|[\?\!]{1,3}|.)/$this->to_tategaki($1,$this->{'rubysize'})/esg);
		$this->{'process'}{'rt'} = '<span style="display:block;position:absolute;left:' . ($this->{'process'}{'fontsize'}[0]+1) . 'px;top:' . (($count == 1) ? (int(($this->{'process'}{'fontsize'}[0] - $this->{'rubysize'}) / 2)) . 'px' : '0') . ';width:' . $this->{'rubysize'} . 'px;font-size:' . $this->{'rubysize'} . 'px;"><span style="display:none">(</span>' . $attribute{'title'} . '<span style="display:none">)</span></span>';
		delete $attribute{'title'};
		$style{'position'} = 'relative';
		$style{'margin'} = '0 auto';
		$style{'width'} = $this->{'process'}{'fontsize'}[0] . 'px';
	}
} elsif ($style{'border'} || $style{'border-left'} || $style{'border-right'}) {
	$style{'margin'} = '0 auto' if !$style{'margin'};
	$style{'width'} = $this->{'process'}{'fontsize'}[0] . 'px' if !$style{'width'};
}
$style{'display'} = 'block';
$attribute{'style'} = (%style) ? $this->hash_to_style(\%style) : '';
return('<' . $$tag . $this->hash_to_attributes(\%attribute) . '>','');
}

sub hash_to_style ($$) {
my($this,$hash) = @_;
my $style = '';
if (ref $hash && %$hash) {
	while (my($k,$v) = each %$hash) {
		next if !defined $k || !defined $v || $k eq '' || $v eq '';
		$style .= $k . ':' . $this->escape($v) . ';';
	}
}
return $style;
}

sub style_to_hash ($$;$) {
my($this,$style,$filter) = @_;
my %hash = ();
if ($style) {
	while ($style =~ /[\x09\x0A\x0C\x0D\x20]*([0-9a-zA-Z_\/\-]+):[\x09\x0A\x0C\x0D\x20]*(.*?)[\x09\x0A\x0C\x0D\x20]*(?:;|$)/isg) {
		my($k,$v) = (lc($1),$2);
		$hash{$k} = $v if !$filter || $k !~ /^(?:$filter)/;
	}
}
return %hash;
}

sub block_close ($;$) {
my($this,$remain) = @_;
$this->reset_inline_property;
shift @{$this->{'process'}{'fontsize'}} if @{$this->{'process'}{'fontsize'}} > 1;
shift @{$this->{'process'}{'line_height'}} if @{$this->{'process'}{'line_height'}} > 1;
shift @{$this->{'process'}{'indent'}};
shift @{$this->{'process'}{'valign'}};
if ($remain) {
	$this->{'process'}{'in_block'}-- if $this->{'process'}{'in_block'};
	$this->{'process'}{'colmargin'} = $this->{'colmargin'} if !$this->{'process'}{'in_block'};
}
}

sub block_open ($$$$$$;$) {
my($this,$line_height,$total_margin_height,$indent,$fontsize_ref,$align,$remain) = @_;
if ($remain) {
	$line_height -= $total_margin_height;
} elsif (!$indent && $total_margin_height) {
	$indent = $total_margin_height;
}
$this->reset_inline_property;
$this->{'process'}{'fontsize'} = [$this->{'fontsize'}];
$this->update_font_size($fontsize_ref) if ref $fontsize_ref && defined $$fontsize_ref && $$fontsize_ref ne '';
unshift(@{$this->{'process'}{'line_height'}},$line_height);
unshift(@{$this->{'process'}{'indent'}},$indent);
unshift(@{$this->{'process'}{'valign'}},(($align && $align eq 'center') ? 'center' : ($align && $align eq 'right') ? 'bottom' : ''));
if ($remain) {
	$this->{'process'}{'in_block'}++;
	$this->{'process'}{'colmargin'} = 0;
}
}

sub reset_inline_property ($) {
my $this = shift;
delete $this->{'process'}{'inline_ltr'};
delete $this->{'process'}{'open_tags'};
delete $this->{'process'}{'prev_open_tags'};
delete $this->{'process'}{'close_tags'};
delete $this->{'process'}{'inline_margin'};
delete $this->{'process'}{'inline_width'};
delete $this->{'process'}{'prev_line_width'};
delete $this->{'process'}{'prev_total_inline_margin'};
}

sub update_font_size ($$) {
my($this,$font_size_ref) = @_;
if (ref $font_size_ref && defined $$font_size_ref) {
	if ($$font_size_ref eq 'xx-small') {
		$$font_size_ref = '9px';
	} elsif ($$font_size_ref eq 'x-small') {
		$$font_size_ref = '11px';
	} elsif ($$font_size_ref eq 'small') {
		$$font_size_ref = '13px';
	} elsif ($$font_size_ref eq 'medium') {
		$$font_size_ref = '16px';
	} elsif ($$font_size_ref eq 'large') {
		$$font_size_ref = '19px';
	} elsif ($$font_size_ref eq 'x-large') {
		$$font_size_ref = '23px';
	} elsif ($$font_size_ref eq 'xx-large') {
		$$font_size_ref = '28px';
	} elsif ($$font_size_ref =~ /([0-9]+\.?[0-9]*)em/) {
		$$font_size_ref = int($this->{'process'}{'fontsize'}[0] * $1) . 'px';
	} elsif ($$font_size_ref =~ /([0-9]+\.?[0-9]*)\%/) {
		$$font_size_ref = int($this->{'process'}{'fontsize'}[0] * $1 / 100) . 'px';
	}
	if ($$font_size_ref && $$font_size_ref =~ /^([0-9]+)px$/) {
		unshift(@{$this->{'process'}{'fontsize'}},$1);
	} else {
		unshift(@{$this->{'process'}{'fontsize'}},$this->{'process'}{'fontsize'}[0]);
		$$font_size_ref = '';
	}
} else {
	unshift(@{$this->{'process'}{'fontsize'}},$this->{'process'}{'fontsize'}[0]);
}
}

sub ratio_to_px ($$$;$$$) {
my($this,$len,$default,$box_height,$opposite_margin,$parent_height) = @_;
$parent_height = $this->{'process'}{'line_height'}[0] if !$parent_height;
if (defined $len && $len eq '0') {
	return 0;
} elsif ($len && $len =~ /([0-9]+)px/) {
	return $1+0;
} elsif ($len && $len =~ /([0-9]+\.?[0-9]*)em/) {
	return ($1) ? int($this->{'process'}{'fontsize'}[0] * $1) : 0;
} elsif ($len && $parent_height && $len =~ /([0-9]+\.?[0-9]*)\%/) {
	return ($1) ? int($parent_height * $1 / 100) : 0;
} elsif ($len && $box_height && defined $opposite_margin && $opposite_margin ne '' && $parent_height && $len eq 'auto') {
	if ($opposite_margin eq 'auto') {
		$len = int(($parent_height - $box_height) / 2);
	} else {
		$len = ($parent_height - $box_height - $this->ratio_to_px($opposite_margin));
	}
	return (($len && $len > 0) ? $len : 0);
} elsif ($default && $default =~ /^[0-9]+\.?[0-9]*$/) {
	return $default;
} else {
	return 0;
}
}

sub num_to_px ($$@) {
my($this,@nums) = @_;
my $v = '';
foreach my $n (@nums) {
	$v .= ((defined $v && $v ne '') ? ' ' : '') . (($n) ? $n . 'px' : '0');
}
return $v;
}

sub calculate_margin ($$;$) {
my($this,$style_hash,$align) = @_;
my($margin_top,$margin_left,$margin_right,$margin_bottom,$padding_top,$padding_left,$padding_right,$padding_bottom,$border_top,$border_left,$border_right,$border_bottom,$box_height);
if ($style_hash->{'padding'}) {
	my @v = split(/[\x09\x0A\x0C\x0D\x20]+/,$style_hash->{'padding'});
	if (@v == 1) {
		$style_hash->{'padding-top'} = $v[0] if !defined $style_hash->{'padding-top'};
		$style_hash->{'padding-right'} = $v[0] if !defined $style_hash->{'padding-right'};
		$style_hash->{'padding-bottom'} = $v[0] if !defined $style_hash->{'padding-bottom'};
		$style_hash->{'padding-left'} = $v[0] if !defined $style_hash->{'padding-left'};
	} elsif (@v == 2) {
		$style_hash->{'padding-top'} = $v[0] if !defined $style_hash->{'padding-top'};
		$style_hash->{'padding-right'} = $v[1] if !defined $style_hash->{'padding-right'};
		$style_hash->{'padding-bottom'} = $v[0] if !defined $style_hash->{'padding-bottom'};
		$style_hash->{'padding-left'} = $v[1] if !defined $style_hash->{'padding-left'};
	} elsif (@v == 3) {
		$style_hash->{'padding-top'} = $v[0] if !defined $style_hash->{'padding-top'};
		$style_hash->{'padding-right'} = $v[1] if !defined $style_hash->{'padding-right'};
		$style_hash->{'padding-bottom'} = $v[2] if !defined $style_hash->{'padding-bottom'};
		$style_hash->{'padding-left'} = $v[1] if !defined $style_hash->{'padding-left'};
	} elsif (@v == 4) {
		$style_hash->{'padding-top'} = $v[0] if !defined $style_hash->{'padding-top'};
		$style_hash->{'padding-right'} = $v[1] if !defined $style_hash->{'padding-right'};
		$style_hash->{'padding-bottom'} = $v[2] if !defined $style_hash->{'padding-bottom'};
		$style_hash->{'padding-left'} = $v[3] if !defined $style_hash->{'padding-left'};
	}
}
$padding_top = $this->ratio_to_px($style_hash->{'padding-left'});
$padding_right = $this->ratio_to_px($style_hash->{'padding-top'});
$padding_bottom = $this->ratio_to_px($style_hash->{'padding-right'});
$padding_left = $this->ratio_to_px($style_hash->{'padding-bottom'});
delete $style_hash->{'padding-top'};
delete $style_hash->{'padding-right'};
delete $style_hash->{'padding-bottom'};
delete $style_hash->{'padding-left'};
if ($padding_top == $padding_right && $padding_top == $padding_bottom && $padding_top == $padding_left) {
	$style_hash->{'padding'} = $this->num_to_px($padding_top);
} elsif ($padding_top == $padding_bottom && $padding_right == $padding_left) {
	$style_hash->{'padding'} = $this->num_to_px($padding_top,$padding_right);
} elsif ($padding_right == $padding_left) {
	$style_hash->{'padding'} = $this->num_to_px($padding_top,$padding_right,$padding_bottom);
} else {
	$style_hash->{'padding'} = $this->num_to_px($padding_top,$padding_right,$padding_bottom,$padding_left);
}

$border_top = (defined $style_hash->{'border-left'}) ? $style_hash->{'border-left'} : '';
$border_right = (defined $style_hash->{'border-top'}) ? $style_hash->{'border-top'} : '';
$border_bottom = (defined $style_hash->{'border-right'}) ? $style_hash->{'border-right'} : '';
$border_left = (defined $style_hash->{'border-bottom'}) ? $style_hash->{'border-bottom'} : '';

$style_hash->{'border-top'} = ($border_top || $style_hash->{'border'}) ? $border_top : 'none';
$style_hash->{'border-right'} = $border_right;
$style_hash->{'border-bottom'} = ($border_bottom || $style_hash->{'border'}) ? $border_bottom : 'none';
$style_hash->{'border-left'} = $border_left;

$border_top = $border_right = $border_bottom = $border_left = $this->ratio_to_px($style_hash->{'border'}) if $style_hash->{'border'};

$border_top = $this->ratio_to_px($style_hash->{'border-top'},$border_top);
$border_right = $this->ratio_to_px($style_hash->{'border-right'},$border_right);
$border_bottom = $this->ratio_to_px($style_hash->{'border-bottom'},$border_bottom);
$border_left = $this->ratio_to_px($style_hash->{'border-left'},$border_left);

if ($style_hash->{'margin'}) {
	my @v = split(/[\x09\x0A\x0C\x0D\x20]+/,$style_hash->{'margin'});
	if (@v == 1) {
		$style_hash->{'margin-top'} = $v[0] if !defined $style_hash->{'margin-top'};
		$style_hash->{'margin-right'} = $v[0] if !defined $style_hash->{'margin-right'};
		$style_hash->{'margin-bottom'} = $v[0] if !defined $style_hash->{'margin-bottom'};
		$style_hash->{'margin-left'} = $v[0] if !defined $style_hash->{'margin-left'};
	} elsif (@v == 2) {
		$style_hash->{'margin-top'} = $v[0] if !defined $style_hash->{'margin-top'};
		$style_hash->{'margin-right'} = $v[1] if !defined $style_hash->{'margin-right'};
		$style_hash->{'margin-bottom'} = $v[0] if !defined $style_hash->{'margin-bottom'};
		$style_hash->{'margin-left'} = $v[1] if !defined $style_hash->{'margin-left'};
	} elsif (@v == 3) {
		$style_hash->{'margin-top'} = $v[0] if !defined $style_hash->{'margin-top'};
		$style_hash->{'margin-right'} = $v[1] if !defined $style_hash->{'margin-right'};
		$style_hash->{'margin-bottom'} = $v[2] if !defined $style_hash->{'margin-bottom'};
		$style_hash->{'margin-left'} = $v[1] if !defined $style_hash->{'margin-left'};
	} elsif (@v == 4) {
		$style_hash->{'margin-top'} = $v[0] if !defined $style_hash->{'margin-top'};
		$style_hash->{'margin-right'} = $v[1] if !defined $style_hash->{'margin-right'};
		$style_hash->{'margin-bottom'} = $v[2] if !defined $style_hash->{'margin-bottom'};
		$style_hash->{'margin-left'} = $v[3] if !defined $style_hash->{'margin-left'};
	}
}
if ($align) {
	if ($align eq 'left') {
		$style_hash->{'margin-right'} = 'auto';
	} elsif ($align eq 'center') {
		$style_hash->{'margin-left'} = $style_hash->{'margin-right'} = 'auto';
	} elsif ($align eq 'right') {
		$style_hash->{'margin-left'} = 'auto';
	}
}
$box_height = $padding_top + $padding_bottom + $border_top + $border_bottom + $1 if $style_hash->{'height'} && $style_hash->{'height'} =~ /([0-9]+)px/;
$margin_top = $this->ratio_to_px($style_hash->{'margin-left'},'',$box_height,$style_hash->{'margin-right'});
$margin_right = $this->ratio_to_px($style_hash->{'margin-top'});
$margin_bottom = $this->ratio_to_px($style_hash->{'margin-right'});
$margin_left = $this->ratio_to_px($style_hash->{'margin-bottom'});
delete $style_hash->{'margin-top'};
delete $style_hash->{'margin-right'};
delete $style_hash->{'margin-bottom'};
delete $style_hash->{'margin-left'};
if ($margin_top == $margin_right && $margin_top == $margin_bottom && $margin_top == $margin_left) {
	$style_hash->{'margin'} = $this->num_to_px($margin_top);
} elsif ($margin_top == $margin_bottom && $margin_right == $margin_left) {
	$style_hash->{'margin'} = $this->num_to_px($margin_top,$margin_right);
} elsif ($margin_right == $margin_left) {
	$style_hash->{'margin'} = $this->num_to_px($margin_top,$margin_right,$margin_bottom);
} else {
	$style_hash->{'margin'} = $this->num_to_px($margin_top,$margin_right,$margin_bottom,$margin_left);
}

return ($margin_left + $margin_right + $padding_left + $padding_right + $border_left + $border_right,$margin_top + $margin_bottom + $padding_top + $padding_bottom + $border_top + $border_bottom);
}

sub add_margin ($$$$$$) {
my($this,$margin,$top,$right,$bottom,$left) = @_;
my($margin_top,$margin_right,$margin_bottom,$margin_left);
my @v = (defined $margin && $margin ne '') ? split(/[\x09\x0A\x0C\x0D\x20]+/,$margin) : (0);
if (@v == 1) {
	$margin_top = $this->ratio_to_px($v[0]) + (($top) ? $top : 0);
	$margin_right = $this->ratio_to_px($v[0]) + (($right) ? $right : 0);
	$margin_bottom = $this->ratio_to_px($v[0]) + (($bottom) ? $bottom : 0);
	$margin_left = $this->ratio_to_px($v[0]) + (($left) ? $left : 0);
} elsif (@v == 2) {
	$margin_top = $this->ratio_to_px($v[0]) + (($top) ? $top : 0);
	$margin_right = $this->ratio_to_px($v[1]) + (($right) ? $right : 0);
	$margin_bottom = $this->ratio_to_px($v[0]) + (($bottom) ? $bottom : 0);
	$margin_left = $this->ratio_to_px($v[1]) + (($left) ? $left : 0);
} elsif (@v == 3) {
	$margin_top = $this->ratio_to_px($v[0]) + (($top) ? $top : 0);
	$margin_right = $this->ratio_to_px($v[1]) + (($right) ? $right : 0);
	$margin_bottom = $this->ratio_to_px($v[2]) + (($bottom) ? $bottom : 0);
	$margin_left = $this->ratio_to_px($v[1]) + (($left) ? $left : 0);
} elsif (@v == 4) {
	$margin_top = $this->ratio_to_px($v[0]) + (($top) ? $top : 0);
	$margin_right = $this->ratio_to_px($v[1]) + (($right) ? $right : 0);
	$margin_bottom = $this->ratio_to_px($v[2]) + (($bottom) ? $bottom : 0);
	$margin_left = $this->ratio_to_px($v[3]) + (($left) ? $left : 0);
}
if ($margin_top == $margin_right && $margin_top == $margin_bottom && $margin_top == $margin_left) {
	return $this->num_to_px($margin_top);
} elsif ($margin_top == $margin_bottom && $margin_right == $margin_left) {
	return $this->num_to_px($margin_top,$margin_right);
} elsif ($margin_right == $margin_left) {
	return $this->num_to_px($margin_top,$margin_right,$margin_bottom);
} else {
	return $this->num_to_px($margin_top,$margin_right,$margin_bottom,$margin_left);
}
}

######################################################################
# 縦書き記号化
sub to_tategaki{
my($this,$w,$fontsize) = @_;

# 制御文字等
if ($w =~ /^[\x00-\x1f]+$/) {
	return $w;

# 半角空白
} elsif ($w =~ /^\x20+$/) {
	$this->{'process'}{'spacer'} = 1;
	return '<span class="latin" style="display:block;width:' . $fontsize . 'px;height:' . $fontsize . 'px;">&nbsp;</span>';

# 回転可能→長音符 ー〜…
} elsif (!$this->{'rotatedisabled'} && $w =~ /[〜ー～∼…]/) {
	return '<span class="cho-on" style="' . (($this->{'css'}) ? '' : 'display:block;margin:auto;-moz-transform:rotate(90deg) scale(1.0,-1.0);-webkit-transform:rotate(90deg) scale(1.0,-1.0);-o-transform:rotate(90deg) scale(1.0,-1.0);-ms-writing-mode:tb-rl;_writing-mode:tb-rl;/writing-mode:tb-rl;') . 'width:' . $fontsize . 'px;height:' . $fontsize . 'px;">' . $w . '</span>';

# 英数字・記号・罫線 || 回転可能→記号；
} elsif ($w =~ /^(&[a-zA-Z0-9]+;|&#[0-9]{2};|[^\x{1100}-\x{11FF}\x{2030}-\x{217F}\x{2460}-\x{24FF}\x{2600}-\x{261B}\x{2620}-\x{277F}\x{2E80}-\x{2FDF}\x{2FF0}-\x{4DBF}\x{4E00}-\x{9FFF}\x{A960}-\x{A97F}\x{AC00}-\x{D7AF}\x{D7B0}-\x{D7FF}\x{F900}-\x{FAFF}\x{FE30}-\x{FE4F}\x{FF00}\x{FF01}\x{FF03}-\x{FF06}\x{FF08}-\x{FF0C}\x{FF0E}-\x{FF1B}\x{FF1F}-\x{FF3D}\x{FF40}-\x{FF5B}\x{FF5D}-\x{FFEF}])$/) {
	return '<span class="latin" style="' . (($this->{'css'}) ? '' : 'display:block;margin:auto;-moz-transform:rotate(90deg);-webkit-transform:rotate(90deg);-o-transform:rotate(90deg);-ms-writing-mode:tb-rl;_writing-mode:tb-rl;/writing-mode:tb-rl;') . 'width:' . $fontsize . 'px;height:' . $fontsize . 'px;">' . $w . '</span>';

# 数字列
} elsif ($w =~ /^[\,\.0-9]+$/) {

# HTMLエンティティ
} elsif ($w =~ /^&(?:#[0-9]+|#x[a-fA-F0-9]+);$/) {

# 変換
} elsif (
		$w =~ s/＜/∧/
	||	$w =~ s/＞/∨/
	||	$w =~ s/↑/→/
	||	$w =~ s/→/↓/
	||	$w =~ s/↓/←/
	||	$w =~ s/←/↑/
	||	$w =~ s/[〜ー～∼]/｜/
	||	$w =~ s/[=＝]/‖/
	||	$w =~ s/…/︙/
	||	$w =~ s/[:;：；]/‥/
	||	$w =~ s/[\-−–]/︲/
	||	$w =~ s/[_＿]/︳/
	||	$w =~ s/[\(（]/︵/
	||	$w =~ s/[\)）]/︶/
	||	$w =~ s/[\{﹛｛]/︷/
	||	$w =~ s/[\}﹜｝]/︸/
	||	$w =~ s/[〔﹝]/︹/
	||	$w =~ s/[〕﹞]/︺/
	||	$w =~ s/【/︻/
	||	$w =~ s/】/︼/
	||	$w =~ s/[«《]/︽/
	||	$w =~ s/[»》]/︾/
	||	$w =~ s/[‹〈〈]/︿/
	||	$w =~ s/[›〉〉]/﹀/
	||	$w =~ s/[「｢]/﹁/
	||	$w =~ s/[」｣]/﹂/
	||	$w =~ s/『/﹃/
	||	$w =~ s/』/﹄/
	||	$w =~ s/[\[［]/﹇/
	||	$w =~ s/[\]］]/﹈/
) {
} elsif ($this->{'lang'} eq 'ja' && (
		$w =~ s/([、，])/'<span class="tou-ten"' . (($this->{'css'}) ? '' : ' style="display:block;margin: -0.7em -0.64em 0.7em 0.64em;position:relative"') . '>' . $1 . '<\/span>'/e		# 、
	||	$w =~ s/([。．])/'<span class="ku-ten"' . (($this->{'css'}) ? '' : ' style="display:block;margin: -0.7em -0.6em 0.7em 0.6em;position:relative"') . '>' . $1 . '<\/span>'/e		# 。
	||	$w =~ s/([ぁぃぅぇぉっゃゅょゎゕゖァィゥェォッャュョヮヵヶ])/'<span class="kogaki"' . (($this->{'css'}) ? '' : ' style="display:block;margin: -0.2em -0.15em 0.2em 0.15em;position:relative"') . '>' . $1 . '<\/span>'/e			#ぁ-ぉっゃ-ょゎ
)) {
	return $w;
}
return '<span class="char"' . (($this->{'css'}) ? '' : ' style="display:block;') . '">' . $w . '</span>';
}

######################################################################
#	(2)-1 obj->text_to_tategaki(text,option);
sub text_to_tategaki ($$;$) {
my($this,$text,$tag) = @_;
$text =~ s/<\!--tate(?![0-9a-zA-Z_])([^>]*)-->(.*?)<\!--\/tate-->/$this->_text_to_tategaki($1,$2,$tag)/egsi;
return $text;
}

sub _text_to_tategaki ($$$;$) {
my($this,$attributes,$text,$tag) = @_;
my $this_tategaki = Taketori->new($this)->set($attributes);
return (($tag) ? '<!--_tate' . $this_tategaki->attributes . ' text="' . $this->escape($text,1) . '"-->' : '') . $this_tategaki->tategaki($text) . (($tag) ? '<!--/_tate-->' : '');
}

sub tategaki_to_text ($$) {
my($this,$text) = @_;
$text =~ s/<\!--_tate(?![0-9a-zA-Z_])([^>]*)-->.*?<\!--\/_tate-->/$this->_tategaki_to_text($1)/egsi;
return $text;
}

sub _tategaki_to_text ($$$;$) {
my($this,$attributes) = @_;
return ($attributes =~ s/[\x09\x0A\x0C\x0D\x20]+text="([^"]+)"//s) ? '<!--tate' . $attributes . '-->' . $this->unescape($1) . '<!--/tate-->' : '<!--tate' . $attributes . '--><!--/tate-->';
}


1;
